package org.nm.challenge.wave;

import java.lang.reflect.InvocationTargetException;

import javax.swing.SwingUtilities;

import org.jfree.ui.RefineryUtilities;
import org.nm.challenge.arrays.Array2D;
import org.nm.challenge.arrays.IArray2D;
import org.nm.challenge.display.CombinedDisplayFrame;
import org.nm.challenge.display.IDisplay;

public class Wave {

	/**
	 * @param args
	 */
	public static void main(String[] args) {
		int debug = 0;
		int nIters = 1000;
		int displayIters = 10;
		long displayWait = 000;
		boolean wavePhysics = true;
		Initialization initType = Initialization.LINEAR_X;
		final int size_y = 500;
		final int size_x = 200;
		Wave wave = new Wave(size_y, size_x, initType, wavePhysics,
				nIters, debug, displayIters, displayWait);
		
		// Start the simulation.
		wave.start();
	}
	
	private final int nIterations;
	private int iteration = 0;
	private final int displayIters;
	private int debug;
	private final IDisplay display;
	private final WaveProblemSetup problemSetup;
	private final long displayWait;
	private final Initialization initType;
	private final int size_y;
	private final int size_x;
	private final double deltaT = 0.05;
	private final double deltaX = 1.0;
	private final double deltaY = 1.0;
	private final double g = 9.80;
	private final Array2D Hx2D;
	private final Array2D Hy2D;
	private final Array2D Ux2D;
	private final Array2D Uy2D;
	private final Array2D Vx2D;
	private final Array2D Vy2D;
	private double H;
	
	public Wave(int size_y, int size_x, Initialization initType,
			boolean wavePhysics, int nIterations, int debug, int displayIters,
			long displayWait) {
		System.out.println(nIterations + " iterations.");
		this.size_y = size_y;
		this.size_x = size_x;
		this.displayWait = displayWait;
		this.nIterations = nIterations;
		this.displayIters = displayIters;
		this.initType = initType;

		this.debug = debug;
		this.display = createDisplay();
		problemSetup = new WaveProblemSetup(size_y, size_x, wavePhysics);
		Hx2D = new Array2D(size_y, size_x+1);
		Ux2D = new Array2D(size_y, size_x+1);
		Vx2D = new Array2D(size_y, size_x+1);
		Hy2D = new Array2D(size_y+1, size_x);
		Uy2D = new Array2D(size_y+1, size_x);
		Vy2D = new Array2D(size_y+1, size_x);
	}

	public int getNIterations() {
		return nIterations;
	}

	/**
	 * Start each worker in its own thread.
	 */
	public void start() {
		long time = System.currentTimeMillis();
		initialize();
		updateDisplay();
		for (int i=0; i<nIterations; ++i) {
			iterate(i);
		}
		long endTime = System.currentTimeMillis();
		System.out.println("done.  Total time: " + (endTime-time));
	}
	
	private void calcTotalH() {
		H = 0.0;
		double[][] dH;
		dH = getH().getData();
		for (int j=1; j<=size_y; ++j) {
			for (int i=1; i<=size_x; ++i) {
				H += dH[j][i];
			}
		}
		if (debug > 0)
			System.out.println("iteration: " + iteration  + " totalH: " + H);
	}

	private IDisplay createDisplay() {
		IDisplay retval = null;
		if (displayIters > 0) {
			//JFreeChartDisplayFrame display = new JFreeChartDisplayFrame();
			//Display2DFrame display = new Display2DFrame();
			CombinedDisplayFrame display = new CombinedDisplayFrame();

			display.pack();
			display.setSize(display.getPreferredSize());
			RefineryUtilities.centerFrameOnScreen(display);
			display.setVisible(true);
			retval = display;
		}
		return retval;
	}

	private void diffusionPhysics(int iter) {
		/* For each element of the matrix ... */
		if (hasDebugOutput()) {System.out.println("Before 1st pass");}
		double[][] dH = getH().getData();;
		double[][] dHNew = getNewH().getData();
	
		double factor = .25;
		for (int j = 1; j <=size_y; j++) {
			for (int i = 1; i<=size_x; i++) {
				dHNew[j][i] = dH[j][i] + factor*((dH[j+1][i]+dH[j-1][i]) +
						(dH[j][i+1]+dH[j][i-1]) - 4.0*dH[j][i]);
			}
		}
	}

	private void initializationCompleted() {
		System.out.println("in initializationCompleted()");
		calcTotalH();
		updateDisplay();
		System.out.println("leaving initializationCompleted()");
	}

	/**
	 * Initialize the 2D arrays
	 * with data.
	 */
	private void initialize() {
		switch (initType) {
		case CIRCLE:
			initializeCircle(50);
			break;
		case GAUSSIAN:
			initializeGaussianCenter(10);
			break;
		case LINEAR_X:
			initializeLinearX();
			break;
		case SQUARE:
			initializeSquare(50);
			break;
		}
		setBoundaryConditions();
		initializationCompleted();
	}

	/**
	 * 
	 */
	private void initializeCircle(int radius) {
		double[][] H = getH().getData();
		double[][] U = getU().getData();
		double[][] V = getV().getData();
	
		int jCenter = size_y / 2;
		int iCenter = size_x / 2;
	
		for (int j=1; j<=size_y; ++j) {
			int dj = Math.abs(j-jCenter);
			for (int i=0; i<=size_x+1; ++i) {
				int di = Math.abs(i-iCenter);
				if (dj*dj + di*di <= radius*radius) {
					H[j][i] = 10.0;
				}
				else {
					H[j][i] = 0;
				}
				U[j][i] = 0.0;
				V[j][i] = 0.0;
			}
		}
		getNewH().copy(getH());
		getNewU().copy(getU());
		getNewV().copy(getV());
	}

	/**
	 * 
	 */
	private void initializeGaussianCenter(double sigma) {
		double[][] H = getH().getData();
		double[][] U = getU().getData();
		double[][] V = getV().getData();
		
		double sigma2 = sigma*sigma;
		int jCenter = size_y / 2;
		int iCenter = size_x / 2;
		
		for (int j=1; j<=size_y; ++j) {
			int dy2 = (j-jCenter)*(j-jCenter);
			for (int i=0; i<=size_x+1; ++i) {
				int dx2 = (i-iCenter)*(i-iCenter);
				H[j][i] = 3 + 7.0*Math.exp(-(dx2+dy2)/sigma2);
				U[j][i] = 0.0;
				V[j][i] = 0.0;
			}
		}
	}

	/**
	 * 
	 */
	private void initializeLinearX() {
		double[][] H = getH().getData();
		double[][] U = getU().getData();
		double[][] V = getV().getData();
		
		final double deltaH = (10.0 - 2.0) / (double)((size_x+1)/2);
		for (int j=1; j<=size_y; ++j) {
			for (int i=0; i<=size_x+1; ++i) {
				H[j][i] = 2.0;
				U[j][i] = 0.0;
				V[j][i] = 0.0;
			}
			
			for (int i=0; i<=(size_x+1)/2; ++i) {
				H[j][i] = 10.0 - (deltaH * (double)i);
			}
		}
	}

	/**
	 * 
	 */
	private void initializeSquare(int radius) {
		double[][] H = getH().getData();
		double[][] U = getU().getData();
		double[][] V = getV().getData();
	
		int jCenter = size_y / 2;
		int iCenter = size_x / 2;
	
		for (int j=1; j<=size_y; ++j) {
			int dj = Math.abs(j-jCenter);
			for (int i=0; i<=size_x+1; ++i) {
				int di = Math.abs(i-iCenter);
				if (dj <= radius && di <= radius) {
					H[j][i] = 10.0;
				}
				else {
					H[j][i] = 0.0;
				}
				U[j][i] = 0.0;
				V[j][i] = 0.0;
			}
		}
	}

	private void iterate(int iter) {
		
		// do the real physics
		if (problemSetup.isWavePhysics()) {
			wavePhysics(iter);
		}
		else {
			diffusionPhysics(iter);
		}
		
		// swap the new array pointers
		// with the current array pointers.
		problemSetup.swapArrays();
		
		// set the boundary conditions on
		// the (freshly swapped) current arrays.
		setBoundaryConditions();
	
		// calculate that portion of the
		// H that lives in the window of
		// the H 2D array that this thread
		// owns.
		calcTotalH();
		
		// The iteration
		// is complete, we can update
		// the display.
		iterationCompleted();
	}

	private void iterationCompleted() {
		++iteration;
		if (displayIters > 0 && (iteration-1) % displayIters == 0) {
			updateDisplay();
		}
	}

	/**
	 * Enforce the boundary conditions
	 */
	private void setBoundaryConditions() {
	
		double H[][] = getH().getData();
		double U[][] = getU().getData();
		double V[][] = getV().getData();
	
		// set boundary conditions
		// on the local part of the window
		// that this thread owns.  This
		// thread owns a slice along
		// the y axis (first index) that
		// spans the entire length (size_x+2) of the
		// x direction (second index)
		// the x boundaries are at index=0
		// and index=size_x+1
		
		for(int j=1; j<=size_y; j++) {
			H[j][0]=H[j][1];
			U[j][0]=-U[j][1];
			V[j][0]=V[j][1];
			H[j][size_x+1]=H[j][size_x];
			U[j][size_x+1]=-U[j][size_x];
			V[j][size_x+1]=V[j][size_x];
		}

		// The y boundary with first index=0
		for(int i=0; i<=size_x+1; i++) {
			H[0][i]=H[1][i];
			U[0][i]=U[1][i];
			V[0][i]=-V[1][i];
		}
		// The y boundary with first index=size_y+1
		for(int i=0; i<=size_x+1; i++) {
			H[size_y+1][i]=H[size_y][i];
			U[size_y+1][i]=U[size_y][i];
			V[size_y+1][i]=-V[size_y][i];
		}
	}

	private double SQ(double d) {
		return d*d;
	}

	private void updateDisplay() {
		if (display == null) {
			return;
		}
		try {
			SwingUtilities.invokeAndWait(new Runnable() {
				public void run() {
					display.update(iteration, problemSetup.getH());
				}
			});
		} catch (InterruptedException e) {
			e.printStackTrace();
		} catch (InvocationTargetException e) {
			e.printStackTrace();
		}
		try {
			Thread.sleep(displayWait);
		} catch (InterruptedException e) {
			e.printStackTrace();
		}
	}

	private void wavePhysics(int iter) {
		   /* For each element of the matrix ... */
	    if (hasDebugOutput()) {System.out.println("Before 1st pass");}
	    double H[][] = getH().getData();
	    double U[][] = getU().getData();
	    double V[][] = getV().getData();
	    double Hx[][] = Hx2D.getData();
	    double Ux[][] = Ux2D.getData();
	    double Vx[][] = Vx2D.getData();
	    //first pass
	    //x direction
	    for (int j = 0; j < size_y; j++) {
	      for (int i = 0; i <= size_x; i++) {
	        //density calculation
	        Hx[j][i]=0.5*(H[j+1][i+1]+H[j+1][i  ]) - deltaT/(2.0*deltaX)*
	                            (U[j+1][i+1]-U[j+1][i  ]);
	        //momentum x calculation
	        Ux[j][i]=0.5*(U[j+1][i+1]   + U[j+1][i  ]) - deltaT/(2.0*deltaX)*
	                           ((SQ(U[j+1][i+1])/H[j+1][i+1] + 0.5*g*SQ(H[j+1][i+1])) -
	                            (SQ(U[j+1][i  ])/H[j+1][i  ] + 0.5*g*SQ(H[j+1][i  ])));
	        //momentum y calculation
	        Vx[j][i]=0.5*(V[j+1][i+1]+V[j+1][i  ]) - deltaT/(2.0*deltaX)*
	                           ((U[j+1][i+1]*V[j+1][i+1]/H[j+1][i+1]) -
	                            (U[j+1][i  ]*V[j+1][i  ]/H[j+1][i  ]));
	        if (debug >= 2) {
	        	System.out.printf("1st pass x direction iter %d i %d j %d %f %f %f %f %f %f %f %f %f\n",iter,i,j,
	              Hx[j][i],Ux[j][i],Vx[j][i],
	              H[j+1][i+1],H[j+1][i],U[j+1][i+1],U[j+1][i],V[j+1][i+1],V[j+1][i]);
	        }
	      }
	    }
	    if (debug >= 2) {
	    	System.out.println();
	    }
	    
	    if(hasDebugOutput()){System.out.println("First pass x direction complete");}
	    
	    double Hy[][] = Hy2D.getData();
	    double Uy[][] = Uy2D.getData();
	    double Vy[][] = Vy2D.getData();
	    //y direction
	    for(int j = 0; j<=size_y; j++){
	      for(int i=0; i<size_x; i++){
	        //density calculation
	        Hy[j][i]=0.5*(H[j+1][i+1]+H[j  ][i+1]) - deltaT/(2.0*deltaY)*
	                            (V[j+1][i+1]-V[j  ][i+1]);
	        //momentum x calculation
	        Uy[j][i]=0.5*(U[j+1][i+1]+U[j  ][i+1]) - deltaT/(2.0*deltaY)*
	                           ((V[j+1][i+1]*U[j+1][i+1]/H[j+1][i+1]) -
	                            (V[j  ][i+1]*U[j  ][i+1]/H[j  ][i+1]));
	        //momentum y calculation
	        Vy[j][i]=0.5*(V[j+1][i+1]  +  V[j  ][i+1]) - deltaT/(2.0*deltaY)*
	                           ((SQ(V[j+1][i+1])/H[j+1][i+1] + 0.5*g*SQ(H[j+1][i+1])) -
	                            (SQ(V[j  ][i+1])/H[j  ][i+1] + 0.5*g*SQ(H[j  ][i+1])));
	        if (debug >= 2) {
	        	System.out.printf("1st pass y direction iter %d i %d j %d %f %f %f %f %f %f %f %f %f\n",iter,i,j,
	              Hy[j][i],Uy[j][i],Vy[j][i],
	              H[j+1][i+1],H[j][i+1],U[j+1][i+1],U[j][i+1],V[j+1][i+1],V[j][i+1]);
	        }
	      }
	    }
	    if (debug >= 2) {
	    	System.out.println();
	    }
	
	    if(hasDebugOutput()){System.out.println("First pass complete");}
	
	    double[][] Hnew = getNewH().getData();
	    double[][] Unew = getNewU().getData();
	    double[][] Vnew = getNewV().getData();
	    
	    //second pass
	    if(hasDebugOutput()){System.out.println("Second Pass started");}
	    for (int j = 1; j <=size_y; j++) {
	      for (int i = 1; i <= size_x; i++) {
	        //density calculation
	        Hnew[j][i] = H[j][i] - (deltaT/deltaX)*(Ux[j-1][i  ]-Ux[j-1][i-1])
	                                   - (deltaT/deltaY)*(Vy[j  ][i-1]-Vy[j-1][i-1]);
	        //momentum x calculation
	        Unew[j][i] = U[j][i] - (deltaT/deltaX)*
	                                    ((SQ(Ux[j-1][i  ])/Hx[j-1][i  ] +0.5*g*SQ(Hx[j-1][i  ])) -
	                                     (SQ(Ux[j-1][i-1])/Hx[j-1][i-1] +0.5*g*SQ(Hx[j-1][i-1])))
	                                   - (deltaT/deltaY)*
	                                    ((Vy[j  ][i-1]*Uy[j  ][i-1]/Hy[j  ][i-1]) -
	                                     (Vy[j-1][i-1]*Uy[j-1][i-1]/Hy[j-1][i-1]));
	        //momentum y calculation
	        Vnew[j][i] = V[j][i] - (deltaT/deltaX)*
	                                    ((Ux[j-1][i  ]*Vx[j-1][i  ]/Hx[j-1][i  ]) -
	                                     (Ux[j-1][i-1]*Vx[j-1][i-1]/Hx[j-1][i-1]))
	                                   - (deltaT/deltaY)*
	                                    ((SQ(Vy[j  ][i-1])/Hy[j  ][i-1] +0.5*g*SQ(Hy[j  ][i-1])) -
	                                     (SQ(Vy[j-1][i-1])/Hy[j-1][i-1] +0.5*g*SQ(Hy[j-1][i-1])));
	        if (debug >= 2) {
	        	System.out.printf("2nd pass iter %d i %d j %d %f %f %f %f %f %f %f %f %f %f\n",iter,i,j,
	              Hnew[j][i],Unew[j][i],Vnew[j][i],
	              H[j][i],U[j][i],V[j][i],Hx[j-1][i-1],Hx[j-1][i],Ux[j-1][i-1],Ux[j-1][i]);
	        }
	      }
	    }
	    if (debug >= 2){
	    	System.out.println();
	    }
	    if(hasDebugOutput()){System.out.println("Second pass complete");} 
	    
	    if(hasDebugOutput()){System.out.println("Done calculations");}
	}

	protected IArray2D getH() {
		return problemSetup.getH();
	}

	protected IArray2D getU() {
		return problemSetup.getU();
	}

	protected IArray2D getV() {
		return problemSetup.getV();
	}

	protected IArray2D getNewH() {
		return problemSetup.getNewH();
	}

	protected IArray2D getNewU() {
		return problemSetup.getNewU();
	}

	protected IArray2D getNewV() {
		return problemSetup.getNewV();
	}

	/**
	 * @return
	 */
	protected boolean hasDebugOutput() {
		return debug>=1;
	}

}
