package org.nm.challenge.wave;

import org.nm.challenge.arrays.Array2D;
import org.nm.challenge.arrays.IArray2D;

public class WaveWorker implements IWaveWorker {

	private final double deltaT = 0.05;
	private final double deltaX = 1.0;
	private final double deltaY = 1.0;
	private final double g = 9.80;
	private double localH = 0.0;
	private final int size_x;
	private final int size_y;
	private final int total_size_y;
	private final int rank;
	private final int debug;
	private SubWaveProblemSetup problemSetup;
	protected final Wave wave;
	private final Initialization initType;
	private final Array2D Hx2D;
	private final Array2D Ux2D;
	private final Array2D Vx2D;
	private final Array2D Hy2D;
	private final Array2D Uy2D;
	private final Array2D Vy2D;

	public WaveWorker(Wave wave, Initialization initType,
			IProblemSetup parentProblemSetup, int rank, int numRanks, int debug) {
		this.wave = wave;
		this.rank = rank;
		this.total_size_y = parentProblemSetup.getTotalSizeY();
		this.problemSetup = new SubWaveProblemSetup(rank, numRanks, parentProblemSetup);
		this.size_x = problemSetup.getSizeX();
		this.size_y = problemSetup.getSizeY();
		this.debug = debug;
		this.initType = initType;
		Hx2D = new Array2D(size_y, size_x+1);
		Ux2D = new Array2D(size_y, size_x+1);
		Vx2D = new Array2D(size_y, size_x+1);
		Hy2D = new Array2D(size_y+1, size_x);
		Uy2D = new Array2D(size_y+1, size_x);
		Vy2D = new Array2D(size_y+1, size_x);
	}

	public int getDebug() {
		return debug;
	}

	public synchronized double getLocalH() {
		return localH;
	}

	public int getRank() {
		return rank;
	}
	
	public int getSize_x() {
		return size_x;
	}

	public int getSize_y() {
		return size_y;
	}

	public Wave getWave() {
		return wave;
	}

	/* (non-Javadoc)
	 * @see java.lang.Runnable#run()
	 */
	public void run() {
		// This run method will be called
		// by the start() method of the
		// Thread object.
		
		
		// initialize the Arrays with data
		initialize();
		
		// enforce boundary conditions
		setBoundaryConditions();
		
		// calculate the total H of the system
		calcLocalH();
		
		// let wave know that we have completed
		// initialization, so it can display
		// the initialized state.
		initializationComplete();
	
		// Loop through the iterations, doing physics
		final int nIterations = wave.getNIterations();
		for (int i = 0; i < nIterations; ++i) {
			try {
				iterate(i);
			} catch (InterruptedException e) {
				return;
			}
		}
	}

	private void calcLocalH() {
		double myH = 0.0;
		double[][] H;
		H = getH().getData();
		for (int j=1; j<=size_y; ++j) {
			for (int i=1; i<=size_x; ++i) {
				myH += H[j][i];
			}
		}
		synchronized(this) {
			localH = myH;
		}
	}

	private void diffusionPhysics(int iter) {
		/* For each element of the matrix ... */
		if (hasDebugOutput()) {System.out.println("Before 1st pass");}
		double[][] dH = getH().getData();;
		double[][] dHNew = getNewH().getData();

		double factor = .25;
		for (int j = 1; j <=size_y; j++) {
			for (int i = 1; i<=size_x; i++) {
				dHNew[j][i] = dH[j][i] + factor*((dH[j+1][i]+dH[j-1][i]) +
						(dH[j][i+1]+dH[j][i-1]) - 4.0*dH[j][i]);
			}
		}
	}

	/**
	 * let's the Wave know that we have completed
	 * initialization, so it can update the display.
	 */
	private void initializationComplete() {
		wave.workerInitCompleted(this);
	}

	/**
	 * Initialize the 2D arrays
	 * with data.
	 */
	private void initialize() {
		switch (initType) {
		case CIRCLE:
			initializeCircle(50);
			break;
		case GAUSSIAN:
			initializeGaussianCenter(10);
			break;
		case LINEAR_X:
			initializeLinearX();
			break;
		case SQUARE:
			initializeSquare(50);
			break;
		}
	}

	/**
	 * 
	 */
	private void initializeCircle(int radius) {
		double[][] H = getH().getData();
		double[][] U = getU().getData();
		double[][] V = getV().getData();
	
		int jCenter = total_size_y / 2;
		int iCenter = size_x / 2;
		int offset_y = problemSetup.getOffsetY();
	
		for (int j=1; j<=size_y; ++j) {
			int jGlobal = j + offset_y;
			int dj = Math.abs(jGlobal-jCenter);
			for (int i=0; i<=size_x+1; ++i) {
				int di = Math.abs(i-iCenter);
				if (dj*dj + di*di <= radius*radius) {
					H[j][i] = 10.0;
				}
				else {
					H[j][i] = 0;
				}
				U[j][i] = 0.0;
				V[j][i] = 0.0;
			}
		}
		getNewH().copy(getH());
		getNewU().copy(getU());
		getNewV().copy(getV());
	}

	/**
	 * 
	 */
	private void initializeGaussianCenter(double sigma) {
		double[][] H = getH().getData();
		double[][] U = getU().getData();
		double[][] V = getV().getData();
		
		double sigma2 = sigma*sigma;
		int jCenter = total_size_y / 2;
		int iCenter = size_x / 2;
		int offset_y = problemSetup.getOffsetY();
		
		for (int j=1; j<=size_y; ++j) {
			int jGlobal = j + offset_y;
			int dy2 = (jGlobal-jCenter)*(jGlobal-jCenter);
			for (int i=0; i<=size_x+1; ++i) {
				int dx2 = (i-iCenter)*(i-iCenter);
				H[j][i] = 3 + 7.0*Math.exp(-(dx2+dy2)/sigma2);
				U[j][i] = 0.0;
				V[j][i] = 0.0;
			}
		}
	}

	/**
	 * 
	 */
	private void initializeLinearX() {
		double[][] H = getH().getData();
		double[][] U = getU().getData();
		double[][] V = getV().getData();
		
		final double deltaH = (10.0 - 2.0) / (double)((size_x+1)/2);
		for (int j=1; j<=size_y; ++j) {
			for (int i=0; i<=size_x+1; ++i) {
				H[j][i] = 2.0;
				U[j][i] = 0.0;
				V[j][i] = 0.0;
			}
			
			for (int i=0; i<=(size_x+1)/2; ++i) {
				H[j][i] = 10.0 - (deltaH * (double)i);
			}
		}
	}

	/**
	 * 
	 */
	private void initializeSquare(int radius) {
		double[][] H = getH().getData();
		double[][] U = getU().getData();
		double[][] V = getV().getData();
	
		int jCenter = total_size_y / 2;
		int iCenter = size_x / 2;
		int offset_y = problemSetup.getOffsetY();
	
		for (int j=1; j<=size_y; ++j) {
			int jGlobal = j + offset_y;
			int dj = Math.abs(jGlobal-jCenter);
			for (int i=0; i<=size_x+1; ++i) {
				int di = Math.abs(i-iCenter);
				if (dj <= radius && di <= radius) {
					H[j][i] = 10.0;
				}
				else {
					H[j][i] = 0.0;
				}
				U[j][i] = 0.0;
				V[j][i] = 0.0;
			}
		}
	}

	private void iterate(int iter) throws InterruptedException {
		
		// do the real physics
		if (problemSetup.isWavePhysics()) {
			wavePhysics(iter);
		}
		else {
			diffusionPhysics(iter);
		}
		
		// swap the new array pointers
		// with the current array pointers.
		problemSetup.swapArrays();
		
		// set the boundary conditions on
		// the (freshly swapped) current arrays.
		setBoundaryConditions();

		// calculate that portion of the
		// H that lives in the window of
		// the H 2D array that this thread
		// owns.
		calcLocalH();
		
		// Tell wave that the iteration
		// is complete, so it can update
		// the display.
		wave.workerIterationCompleted(this);
	}

	/**
	 * Enforce the boundary conditions
	 */
	private void setBoundaryConditions() {
	
		double H[][] = getH().getData();
		double U[][] = getU().getData();
		double V[][] = getV().getData();
	
		// set boundary conditions
		// on the local part of the window
		// that this thread owns.  This
		// thread owns a slice along
		// the y axis (first index) that
		// spans the entire length (size_x+2) of the
		// x direction (second index)
		// the x boundaries are at index=0
		// and index=size_x+1
		
		for(int j=1; j<=size_y; j++) {
			H[j][0]=H[j][1];
			U[j][0]=-U[j][1];
			V[j][0]=V[j][1];
			H[j][size_x+1]=H[j][size_x];
			U[j][size_x+1]=-U[j][size_x];
			V[j][size_x+1]=V[j][size_x];
		}
		
		// If this thread is the first thread,
		// then it is also responsible for the
		// y boundary with first index=0
		if (rank == 0) {
			for(int i=0; i<=size_x+1; i++) {
				H[0][i]=H[1][i];
				U[0][i]=U[1][i];
				V[0][i]=-V[1][i];
			}
		}
		int numRanks = problemSetup.getNumRanks();
		// If this thread is the last thread,
		// then it is also responsible for the
		// y boundary with first index=size_y+1
		if (rank == numRanks-1) {
			for(int i=0; i<=size_x+1; i++) {
				H[size_y+1][i]=H[size_y][i];
				U[size_y+1][i]=U[size_y][i];
				V[size_y+1][i]=-V[size_y][i];
			}
		}
	}

	private double SQ(double d) {
		return d*d;
	}

	private void wavePhysics(int iter) {
		   /* For each element of the matrix ... */
	    if (hasDebugOutput()) {System.out.println("Before 1st pass");}
	    double H[][] = getH().getData();
	    double U[][] = getU().getData();
	    double V[][] = getV().getData();
	    double Hx[][] = Hx2D.getData();
	    double Ux[][] = Ux2D.getData();
	    double Vx[][] = Vx2D.getData();
	    //first pass
	    //x direction
	    for (int j = 0; j < size_y; j++) {
	      for (int i = 0; i <= size_x; i++) {
	        //density calculation
	        Hx[j][i]=0.5*(H[j+1][i+1]+H[j+1][i  ]) - deltaT/(2.0*deltaX)*
	                            (U[j+1][i+1]-U[j+1][i  ]);
	        //momentum x calculation
	        Ux[j][i]=0.5*(U[j+1][i+1]   + U[j+1][i  ]) - deltaT/(2.0*deltaX)*
	                           ((SQ(U[j+1][i+1])/H[j+1][i+1] + 0.5*g*SQ(H[j+1][i+1])) -
	                            (SQ(U[j+1][i  ])/H[j+1][i  ] + 0.5*g*SQ(H[j+1][i  ])));
	        //momentum y calculation
	        Vx[j][i]=0.5*(V[j+1][i+1]+V[j+1][i  ]) - deltaT/(2.0*deltaX)*
	                           ((U[j+1][i+1]*V[j+1][i+1]/H[j+1][i+1]) -
	                            (U[j+1][i  ]*V[j+1][i  ]/H[j+1][i  ]));
	        if (debug >= 2) {
	        	System.out.printf("1st pass x direction iter %d i %d j %d %f %f %f %f %f %f %f %f %f\n",iter,i,j,
	              Hx[j][i],Ux[j][i],Vx[j][i],
	              H[j+1][i+1],H[j+1][i],U[j+1][i+1],U[j+1][i],V[j+1][i+1],V[j+1][i]);
	        }
	      }
	    }
	    if (debug >= 2) {
	    	System.out.println();
	    }
	    
	    if(hasDebugOutput()){System.out.println("First pass x direction complete");}
	    
	    double Hy[][] = Hy2D.getData();
	    double Uy[][] = Uy2D.getData();
	    double Vy[][] = Vy2D.getData();
	    //y direction
	    for(int j = 0; j<=size_y; j++){
	      for(int i=0; i<size_x; i++){
	        //density calculation
	        Hy[j][i]=0.5*(H[j+1][i+1]+H[j  ][i+1]) - deltaT/(2.0*deltaY)*
	                            (V[j+1][i+1]-V[j  ][i+1]);
	        //momentum x calculation
	        Uy[j][i]=0.5*(U[j+1][i+1]+U[j  ][i+1]) - deltaT/(2.0*deltaY)*
	                           ((V[j+1][i+1]*U[j+1][i+1]/H[j+1][i+1]) -
	                            (V[j  ][i+1]*U[j  ][i+1]/H[j  ][i+1]));
	        //momentum y calculation
	        Vy[j][i]=0.5*(V[j+1][i+1]  +  V[j  ][i+1]) - deltaT/(2.0*deltaY)*
	                           ((SQ(V[j+1][i+1])/H[j+1][i+1] + 0.5*g*SQ(H[j+1][i+1])) -
	                            (SQ(V[j  ][i+1])/H[j  ][i+1] + 0.5*g*SQ(H[j  ][i+1])));
	        if (debug >= 2) {
	        	System.out.printf("1st pass y direction iter %d i %d j %d %f %f %f %f %f %f %f %f %f\n",iter,i,j,
	              Hy[j][i],Uy[j][i],Vy[j][i],
	              H[j+1][i+1],H[j][i+1],U[j+1][i+1],U[j][i+1],V[j+1][i+1],V[j][i+1]);
	        }
	      }
	    }
	    if (debug >= 2) {
	    	System.out.println();
	    }
	
	    if(hasDebugOutput()){System.out.println("First pass complete");}
	
	    double[][] Hnew = getNewH().getData();
	    double[][] Unew = getNewU().getData();
	    double[][] Vnew = getNewV().getData();
	    
	    //second pass
	    if(hasDebugOutput()){System.out.println("Second Pass started");}
	    for (int j = 1; j <=size_y; j++) {
	      for (int i = 1; i <= size_x; i++) {
	        //density calculation
	        Hnew[j][i] = H[j][i] - (deltaT/deltaX)*(Ux[j-1][i  ]-Ux[j-1][i-1])
	                                   - (deltaT/deltaY)*(Vy[j  ][i-1]-Vy[j-1][i-1]);
	        //momentum x calculation
	        Unew[j][i] = U[j][i] - (deltaT/deltaX)*
	                                    ((SQ(Ux[j-1][i  ])/Hx[j-1][i  ] +0.5*g*SQ(Hx[j-1][i  ])) -
	                                     (SQ(Ux[j-1][i-1])/Hx[j-1][i-1] +0.5*g*SQ(Hx[j-1][i-1])))
	                                   - (deltaT/deltaY)*
	                                    ((Vy[j  ][i-1]*Uy[j  ][i-1]/Hy[j  ][i-1]) -
	                                     (Vy[j-1][i-1]*Uy[j-1][i-1]/Hy[j-1][i-1]));
	        //momentum y calculation
	        Vnew[j][i] = V[j][i] - (deltaT/deltaX)*
	                                    ((Ux[j-1][i  ]*Vx[j-1][i  ]/Hx[j-1][i  ]) -
	                                     (Ux[j-1][i-1]*Vx[j-1][i-1]/Hx[j-1][i-1]))
	                                   - (deltaT/deltaY)*
	                                    ((SQ(Vy[j  ][i-1])/Hy[j  ][i-1] +0.5*g*SQ(Hy[j  ][i-1])) -
	                                     (SQ(Vy[j-1][i-1])/Hy[j-1][i-1] +0.5*g*SQ(Hy[j-1][i-1])));
	        if (debug >= 2) {
	        	System.out.printf("2nd pass iter %d i %d j %d %f %f %f %f %f %f %f %f %f %f\n",iter,i,j,
	              Hnew[j][i],Unew[j][i],Vnew[j][i],
	              H[j][i],U[j][i],V[j][i],Hx[j-1][i-1],Hx[j-1][i],Ux[j-1][i-1],Ux[j-1][i]);
	        }
	      }
	    }
	    if (debug >= 2){
	    	System.out.println();
	    }
	    if(hasDebugOutput()){System.out.println("Second pass complete");} 
	    
	    if(hasDebugOutput()){System.out.println("Done calculations");}
	}

	protected IArray2D getH() {
		return problemSetup.getH();
	}

	protected IArray2D getU() {
		return problemSetup.getU();
	}

	protected IArray2D getV() {
		return problemSetup.getV();
	}

	protected IArray2D getNewH() {
		return problemSetup.getNewH();
	}

	protected IArray2D getNewU() {
		return problemSetup.getNewU();
	}

	protected IArray2D getNewV() {
		return problemSetup.getNewV();
	}

	/**
	 * @return
	 */
	protected boolean hasDebugOutput() {
		return rank == 0 && debug>=1;
	}

}