#include <mpi.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <unistd.h>
#include "wave.h"

/*********************************************************************************************
 * WAVE -- 2D Shallow Water Equation Model
 *                            Bob Robey, Los Alamos National Laboratory
 *                      Written for Supercomputing Challenge Kickoff 2007
 *                                   Copyright, 2007
 *                      Permission Granted to Supercomputing Challenge
 *                           Participants to use in their projects
 
 * This work is based on the following 2006-2007 Supercomputing
 * Challenge Projects. Their work is used by permission.
 *  
 * Numerical method based on the Shallow Water Model by
 * 
 *             Joseph Koby, Sarah Armstrong, Juan-Antonio Vigil, Vanessa Trujillo,
 *                                  McCurdy High School
 *                                    Copyright 2007
 *                        Written for Supercomputing Challenge 2006-2007
 *                                ALL RIGHTS RESERVED
 * 
 * C/MPI coding based on the SAPIENT Compressible Fluid Model
 * 
 *                                      SAPIENT
 *                  Copyright 2007 Jonathan Robey, Dov Shlachter
 *                               Los Alamos High School
 *                     Written for Supercomputing Challenge 2006-2007
 *                                  ALL RIGHTS RESERVED
 * 
 * ******************************************************************************************/

//define macro for squaring a number
#define SQ(x) ((x)*(x))

/* Display routines */
void display_init(char *displayname, int iwidth, int iheight);
void display_one_d(int matrix_size_y, int matrix_size_x, double **temp, int my_offset,
  int mysize, double maxscale);
void set_label(char *text);
void display_close(void);

/* Memory allocation routines */
double *dvector(int n);
double **dmatrix(int m, int n);

double   **H, **U, **V;                 //state variables
double   **Hnew, **Unew, **Vnew;        //
double   **Hx, **Ux, **Vx;  //half-step arrays
double   **Hy, **Uy, **Vy;  //half-step arrays

int main(int argc, char *argv[])
{
  int      rank, size ;
  int      next, prev ;
  int      i, j;
  int      matrix_size_x, matrix_size_y;
  int      ntimes;
  int      n;
  int      mysize;
  int      my_offset;
  double   deltaX, deltaY;         //size of cell
  double   g = 9.80;                           // gravitational constant
  double   **temp;                             //holder array for display
  double   deltaT = 0.05;                       //hardwired timestep
  double   maxScale;                           //
  double   time=0.0;                           //computer simulation time
  double   slavetime, totaltime, starttime ;   //variables to calculate time taken for the program to run
  double   myTM, TotalMass, origTM;    //variables for checking conservation of mass
  char     *desc;           //variable for labels
  char     string[80];
  char     *displayname = ":0";
    
  MPI_Init (&argc, &argv);
  
  /* Determine size and my rank in MPI_COMM_WORLD communicator */
  
  MPI_Comm_size(MPI_COMM_WORLD, &size) ;
  MPI_Comm_rank(MPI_COMM_WORLD, &rank) ;    

  if (argc > 2 && strcmp( argv[1], "-display" ) == 0) {
      displayname = (char *)malloc( strlen( argv[2] ) + 1 );
      strcpy(displayname, argv[2]);
  }
    
  if(display_on)  display_init(displayname, 1000, 400);

  if(rank==0){printf("Copyright 2007\n");}
 
    /* If I'm process 0, determine the matrix size and # of iterations*/ 
  if ( rank == 0) {
      /*
      printf("Matrix Size X : ");
      scanf("%d",&matrix_size_x);
      printf("Matrix Size Y : ") ;
      scanf("%d",&matrix_size_y) ;
      printf("Iterations : ") ;
      scanf("%d",&ntimes) ;//*/
      matrix_size_x = 500; matrix_size_y = 200; ntimes =2000;
  }
  
  /* Broadcast the size and # of itertations to all processes */
  MPI_Bcast(&matrix_size_x, 1, MPI_INT, 0, MPI_COMM_WORLD) ;
  MPI_Bcast(&matrix_size_y, 1, MPI_INT, 0, MPI_COMM_WORLD) ;
  MPI_Bcast(&ntimes, 1, MPI_INT, 0, MPI_COMM_WORLD);

  /* Set neighbors */
  if (rank == 0)
    prev = MPI_PROC_NULL;
  else
    prev = rank-1;
  if (rank == size - 1)
    next = MPI_PROC_NULL;
  else
    next = rank+1;

  /* Determine my part of the matrix */
  mysize = matrix_size_y/size + ((rank < (matrix_size_y % size)) ? 1 : 0 ) ;
  my_offset = rank * (matrix_size_y/size);
  if (rank > (matrix_size_y % size)) my_offset += (matrix_size_y % size);
  else                               my_offset += rank;

  if(debug>=1){printf("my rank is %d and mysize is %d\n", rank,mysize);}
  /* allocate the memory dynamically for the matrix */
  H = dmatrix(mysize+2, matrix_size_x+2);
  U = dmatrix(mysize+2, matrix_size_x+2);
  V = dmatrix(mysize+2, matrix_size_x+2);

  Hnew = dmatrix(mysize+2, matrix_size_x+2);
  Unew = dmatrix(mysize+2, matrix_size_x+2);
  Vnew = dmatrix(mysize+2, matrix_size_x+2);

  Hx = dmatrix(mysize, matrix_size_x+1);
  Ux = dmatrix(mysize, matrix_size_x+1);
  Vx = dmatrix(mysize, matrix_size_x+1);

  Hy = dmatrix(mysize+1, matrix_size_x);
  Uy = dmatrix(mysize+1, matrix_size_x);
  Vy = dmatrix(mysize+1, matrix_size_x);

  temp = dmatrix(mysize+2, matrix_size_x+2);
   
  if(rank==0&&debug>=1){printf("Memory allocated\n");}
  /*initialize matrix*/
  
  for(j=0;j<=mysize+1;j++){
    for(i=0;i<=matrix_size_x+1;i++){
      H[j][i]=2.0;
      U[j][i]=0.0;
      V[j][i]=0.0;
    }
      for(i=0;i<=(matrix_size_x+1)/2;i++){
        H[j][i]=10.0 - ((10.0 - 2.0)/ (double)((matrix_size_x+1)/2))*(double)(i);
      }
  }
  deltaX=1.0;
  deltaY=1.0;
  
  if(rank==0&&debug>=1){printf("initial values set\n");}
  //display initial values
 
  if (display_on) { 
    maxScale=10.0;
    desc="initial values of height";
    set_label(desc);
    display_one_d(matrix_size_x, matrix_size_y, H, my_offset, mysize, maxScale);
  }

  myTM=0.0;
  for(j=1;j<=mysize;j++){
    for(i=1;i<=matrix_size_x;i++){
        myTM+=H[j][i];
    }
  }
  MPI_Allreduce(&myTM, &origTM, 1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
  
  if(rank==0&&debug>=1){printf("initial values displayed\n");}
  /* run the simulation for given number of iterations */
  starttime = MPI_Wtime() ;
  for (n = 0; n < ntimes; n++) {
    MPI_Request      req[12];
    MPI_Status       status[12];

    /* Send and receive boundary information */
    MPI_Isend(H[1],matrix_size_x+2,MPI_DOUBLE,prev,1,MPI_COMM_WORLD,req); 
    MPI_Irecv(H[mysize+1],matrix_size_x+2,MPI_DOUBLE,next,1,MPI_COMM_WORLD,req+1);

    MPI_Isend(H[mysize],matrix_size_x+2,MPI_DOUBLE,next,2,MPI_COMM_WORLD,req+2);
    MPI_Irecv(H[0],matrix_size_x+2,MPI_DOUBLE,prev,2,MPI_COMM_WORLD,req+3);

    if(rank==0&&debug>=1){printf("values for H communicated\n");}
    
    MPI_Isend(U[1],matrix_size_x+2,MPI_DOUBLE,prev,3,MPI_COMM_WORLD,req+4); 
    MPI_Irecv(U[mysize+1],matrix_size_x+2,MPI_DOUBLE,next,3,MPI_COMM_WORLD,req+5);

    MPI_Isend(U[mysize],matrix_size_x+2,MPI_DOUBLE,next,4,MPI_COMM_WORLD,req+6);
    MPI_Irecv(U[0],matrix_size_x+2,MPI_DOUBLE,prev,4,MPI_COMM_WORLD,req+7);

    if(rank==0&&debug>=1){printf("values for x momentum communicated\n");}

    MPI_Isend(V[1],matrix_size_x+2,MPI_DOUBLE,prev,5,MPI_COMM_WORLD,req+8); 
    MPI_Irecv(V[mysize+1],matrix_size_x+2,MPI_DOUBLE,next,5,MPI_COMM_WORLD,req+9);

    MPI_Isend(V[mysize],matrix_size_x+2,MPI_DOUBLE,next,6,MPI_COMM_WORLD,req+10);
    MPI_Irecv(V[0],matrix_size_x+2,MPI_DOUBLE,prev,6,MPI_COMM_WORLD,req+11);

    if(rank==0&&debug>=1){printf("values for y momentum communicated\n");}
    MPI_Waitall(12, req, status);

    if(rank==0&&debug>=1)printf("Communication successful\n");
    //set boundary conditons
    for(j=1;j<=mysize;j++){
      H[j][0]=H[j][1];
      U[j][0]=-U[j][1];
      V[j][0]=V[j][1];
      H[j][matrix_size_x+1]=H[j][matrix_size_x];
      U[j][matrix_size_x+1]=-U[j][matrix_size_x];
      V[j][matrix_size_x+1]=V[j][matrix_size_x];
      }
    for(i=0;i<=matrix_size_x+1;i++){
      if(my_offset==0){
        H[0][i]=H[1][i];
        U[0][i]=U[1][i];
        V[0][i]=-V[1][i];
      }
      if(matrix_size_y==my_offset+mysize){
        H[mysize+1][i]=H[mysize][i];
        U[mysize+1][i]=U[mysize][i];
        V[mysize+1][i]=-V[mysize][i];
      }
    }
    if(rank==0&&debug>=1)printf("Boundary conditions set\n");
    
    //set timestep
    // We should calculate a timestep here based on SQ(c)=g*h and not let a wave
    // cross a cell. But it does not vary much so a constant timestep is used for now

    time+=deltaT;
    if(rank==0&&debug>=1){printf("deltaT set to %20.1f\n",deltaT);}

    /* For each element of the matrix ... */
    if (rank == 0&&debug>=1) {printf("Before 1st pass\n");}
    //first pass
    //x direction
    for (j = 0; j < mysize; j++) {
      for (i = 0; i<=matrix_size_x;i++) {
        //density calculation
        Hx[j][i]=0.5*(H[j+1][i+1]+H[j+1][i  ]) - deltaT/(2.0*deltaX)*
                            (U[j+1][i+1]-U[j+1][i  ]);
        //momentum x calculation
        Ux[j][i]=0.5*(U[j+1][i+1]+U[j+1][i  ]) - deltaT/(2.0*deltaX)*
                           ((SQ(U[j+1][i+1])/H[j+1][i+1] + 0.5*g*SQ(H[j+1][i+1])) -
                            (SQ(U[j+1][i  ])/H[j+1][i  ] + 0.5*g*SQ(H[j+1][i  ])));
        //momentum y calculation
        Vx[j][i]=0.5*(V[j+1][i+1]+V[j+1][i  ]) - deltaT/(2.0*deltaX)*
                           ((U[j+1][i+1]*V[j+1][i+1]/H[j+1][i+1]) -
                            (U[j+1][i  ]*V[j+1][i  ]/H[j+1][i  ]));
        if (debug >= 2) {
          printf("1st pass x direction iter %d i %d j %d %lf %lf %lf %lf %lf %lf %lf %lf %lf\n",n,i,j,
              Hx[j][i],Ux[j][i],Vx[j][i],
              H[j+1][i+1],H[j+1][i],U[j+1][i+1],U[j+1][i],V[j+1][i+1],V[j+1][i]);
        }
      }
    }
    if (debug >= 2) {
      printf("\n");
    }
    
    if(rank==0&&debug>=1){printf("First pass x direction complete\n");}
    
    //y direction
    for(j = 0; j<=mysize; j++){
      for(i=0; i<matrix_size_x; i++){
        //density calculation
        Hy[j][i]=0.5*(H[j+1][i+1]+H[j  ][i+1]) - deltaT/(2.0*deltaY)*
                            (V[j+1][i+1]-V[j  ][i+1]);
        //momentum x calculation
        Uy[j][i]=0.5*(U[j+1][i+1]+U[j  ][i+1]) - deltaT/(2.0*deltaY)*
                           ((V[j+1][i+1]*U[j+1][i+1]/H[j+1][i+1]) -
                            (V[j  ][i+1]*U[j  ][i+1]/H[j  ][i+1]));
        //momentum y calculation
        Vy[j][i]=0.5*(V[j+1][i+1]+V[j  ][i+1]) - deltaT/(2.0*deltaY)*
                           ((SQ(V[j+1][i+1])/H[j+1][i+1] + 0.5*g*SQ(H[j+1][i+1])) -
                            (SQ(V[j  ][i+1])/H[j  ][i+1] + 0.5*g*SQ(H[j  ][i+1])));
        if (debug >= 2) {
          printf("1st pass y direction iter %d i %d j %d %lf %lf %lf %lf %lf %lf %lf %lf %lf\n",n,i,j,
              Hy[j][i],Uy[j][i],Vy[j][i],
              H[j+1][i+1],H[j][i+1],U[j+1][i+1],U[j][i+1],V[j+1][i+1],V[j][i+1]);
        }
      }
    }
    if (debug >= 2) {
      printf("\n");
    }

    if(rank==0&&debug>=1){printf("First pass complete\n");}

    //second pass
    if(rank==0&&debug>=1){printf("Second Pass started\n");}
    for (j = 1; j <=mysize; j++) {
      for (i = 1; i<=matrix_size_x;i++) {
        //density calculation
        Hnew[j][i] = H[j][i] - (deltaT/deltaX)*(Ux[j-1][i  ]-Ux[j-1][i-1])
                                   - (deltaT/deltaY)*(Vy[j  ][i-1]-Vy[j-1][i-1]);
        //momentum x calculation
        Unew[j][i] = U[j][i] - (deltaT/deltaX)*
                                ((SQ(Ux[j-1][i  ])/Hx[j-1][i  ] +0.5*g*SQ(Hx[j-1][i  ])) -
                                 (SQ(Ux[j-1][i-1])/Hx[j-1][i-1] +0.5*g*SQ(Hx[j-1][i-1])))
                             - (deltaT/deltaY)*
                                ((Vy[j  ][i-1]*Uy[j  ][i-1]/Hy[j  ][i-1]) -
                                 (Vy[j-1][i-1]*Uy[j-1][i-1]/Hy[j-1][i-1]));
        //momentum y calculation
        Vnew[j][i] = V[j][i] - (deltaT/deltaX)*
                                ((Ux[j-1][i  ]*Vx[j-1][i  ]/Hx[j-1][i  ]) -
                                 (Ux[j-1][i-1]*Vx[j-1][i-1]/Hx[j-1][i-1]))
                             - (deltaT/deltaY)*
                                ((SQ(Vy[j  ][i-1])/Hy[j  ][i-1] +0.5*g*SQ(Hy[j  ][i-1])) -
                                 (SQ(Vy[j-1][i-1])/Hy[j-1][i-1] +0.5*g*SQ(Hy[j-1][i-1])));
        if (debug >= 2) {
          printf("2nd pass iter %d i %d j %d %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf\n",
              n,i,j,Hnew[j][i],Unew[j][i],Vnew[j][i],
              H[j][i],U[j][i],V[j][i],Hx[j-1][i-1],Hx[j-1][i],Ux[j-1][i-1],Ux[j-1][i]);
        }
      }
    }
    if (debug >= 2){
      printf("\n");
    }
    if(rank==0&&debug>=1){printf("Second pass complete\n");} 
    
    if(rank==0&&debug>=1){printf("Done calculations\n");}
    //set temp to ____ and display
    for (j = 0; j <=mysize+1; j++) {
      for (i = 0; i<=matrix_size_x+1;i++) {
        H[j][i] = Hnew[j][i];
        U[j][i] = Unew[j][i];
        V[j][i] = Vnew[j][i];
      }
    }
    if (debug >= 2) {
      for(j=0;j<=mysize+1;j++){
        for(i=0;i<=matrix_size_x+1;i++){
          printf("end of cycle %d %d %lf %lf %lf\n",i,j,H[j][i],U[j][i], V[j][i]);
        }
      }
    }  
    if (display_on){
      maxScale=10.0;
      desc="height";//mass or height
      sprintf(string,"%s     iter %d      time %lf",desc,n,time);
      set_label(string);
      if (n%display_on == 0){
        display_one_d(matrix_size_x, matrix_size_y, H, my_offset, mysize, maxScale);
      }
    }
      
    myTM=0.0;
    TotalMass=0.0;
    for(j=1;j<=mysize;j++){
      for(i=1;i<=matrix_size_x;i++){
          if (isnan(H[j][i])) {
            printf("Error -- H[%d][%d]=%f\n",i,j,H[j][i]);
          }
          myTM+=H[j][i];
      }
    }
    MPI_Allreduce(&myTM, &TotalMass, 1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);

    if(((fabs(TotalMass-origTM)>1.0E-6)||isnan(TotalMass))&&check==1){
       printf("Conservation of mass\nMass difference:%e\n", TotalMass-origTM);
       printf("Problem occured on iteration %5.5d at time %f.\n", n, time);
       exit(0);
    }
    //print iteration info
    if (display_on && n%display_on == 0) {
      if (rank == 0){
        printf("Iteration:%5.5d, Time:%f, Timestep:%f Total mass:%f\n", n, time, deltaT, TotalMass);
      }
    }

  }  // End of iteration loop
  
  /* Compute the average time taken/processor */
  slavetime = MPI_Wtime() - starttime;
  MPI_Reduce (&slavetime, &totaltime, 1, MPI_DOUBLE, MPI_SUM, 0, MPI_COMM_WORLD);
  
  /* Print the total time taken */
  if (rank == 0)
    printf("[%d] Flow finished in %lf seconds\n",rank,totaltime/(double)size);//*/

  if(display_on)
    display_close();
    
  // Should free memory allocated with dmatrix call
  
  MPI_Finalize();
  exit(0);
}
